/**
 * Copyright (C) 2012 - present by OpenGamma Inc. and the OpenGamma group of companies
 * 
 * Please see distribution for license.
 */
package com.opengamma.analytics.financial.curve; 

import static org.testng.AssertJUnit.assertEquals; 
import static org.threeten.bp.temporal.ChronoUnit.DAYS; 
import static org.threeten.bp.temporal.ChronoUnit.MONTHS; 
import static org.threeten.bp.temporal.ChronoUnit.YEARS; 

import java.util.ArrayList; 
import java.util.Arrays; 
import java.util.HashMap; 
import java.util.List; 
import java.util.Set; 

import org.testng.annotations.BeforeSuite; 
import org.testng.annotations.Test; 
import org.threeten.bp.LocalDate; 
import org.threeten.bp.Period; 
import org.threeten.bp.ZonedDateTime; 

import com.opengamma.analytics.financial.calculator.MarketQuoteSensitivityBlockCalculator; 
import com.opengamma.analytics.financial.calculator.PresentValueConvertedCalculator; 
import com.opengamma.analytics.financial.calculator.PresentValueCurveSensitivityConvertedCalculator; 
import com.opengamma.analytics.financial.calculator.PresentValueCurveSensitivityMCSCalculator; 
import com.opengamma.analytics.financial.calculator.PresentValueMCACalculator; 
import com.opengamma.analytics.financial.curve.building.CurveBuildingBlockBundle; 
import com.opengamma.analytics.financial.curve.building.CurveBuildingFunction; 
import com.opengamma.analytics.financial.curve.generator.GeneratorCurveAddYield; 
import com.opengamma.analytics.financial.curve.generator.GeneratorCurveAddYieldExisiting; 
import com.opengamma.analytics.financial.curve.generator.GeneratorCurveAddYieldFixed; 
import com.opengamma.analytics.financial.curve.generator.GeneratorCurveAddYieldNb; 
import com.opengamma.analytics.financial.curve.generator.GeneratorCurveDiscountFactorInterpolated; 
import com.opengamma.analytics.financial.curve.generator.GeneratorCurveYieldInterpolated; 
import com.opengamma.analytics.financial.curve.generator.GeneratorCurveYieldInterpolatedAnchor; 
import com.opengamma.analytics.financial.curve.generator.GeneratorCurveYieldNelsonSiegel; 
import com.opengamma.analytics.financial.curve.generator.GeneratorCurveYieldPeriodicInterpolated; 
import com.opengamma.analytics.financial.curve.generator.GeneratorYDCurve; 
import com.opengamma.analytics.financial.curve.sensitivity.ParameterSensitivity; 
import com.opengamma.analytics.financial.curve.sensitivity.ParameterUnderlyingSensitivityBlockCalculator; 
import com.opengamma.analytics.financial.forex.method.FXMatrix; 
import com.opengamma.analytics.financial.forex.method.MultipleCurrencyInterestRateCurveSensitivity; 
import com.opengamma.analytics.financial.instrument.InstrumentDefinition; 
import com.opengamma.analytics.financial.instrument.cash.CashDefinition; 
import com.opengamma.analytics.financial.instrument.fra.ForwardRateAgreementDefinition; 
import com.opengamma.analytics.financial.instrument.index.GeneratorDeposit; 
import com.opengamma.analytics.financial.instrument.index.GeneratorDepositON; 
import com.opengamma.analytics.financial.instrument.index.GeneratorInstrument; 
import com.opengamma.analytics.financial.instrument.index.GeneratorSwapFixedIbor; 
import com.opengamma.analytics.financial.instrument.index.GeneratorSwapFixedIborMaster; 
import com.opengamma.analytics.financial.instrument.index.GeneratorSwapFixedON; 
import com.opengamma.analytics.financial.instrument.index.IndexON; 
import com.opengamma.analytics.financial.instrument.swap.SwapFixedIborDefinition; 
import com.opengamma.analytics.financial.instrument.swap.SwapFixedONDefinition; 
import com.opengamma.analytics.financial.interestrate.InstrumentDerivative; 
import com.opengamma.analytics.financial.interestrate.InstrumentDerivativeVisitor; 
import com.opengamma.analytics.financial.interestrate.InterestRateCurveSensitivity; 
import com.opengamma.analytics.financial.interestrate.LastTimeCalculator; 
import com.opengamma.analytics.financial.interestrate.ParSpreadMarketQuoteCalculator; 
import com.opengamma.analytics.financial.interestrate.ParSpreadMarketQuoteCurveSensitivityCalculator; 
import com.opengamma.analytics.financial.interestrate.YieldCurveBundle; 
import com.opengamma.analytics.financial.interestrate.payments.derivative.Coupon; 
import com.opengamma.analytics.financial.interestrate.swap.derivative.SwapFixedCoupon; 
import com.opengamma.analytics.financial.model.interestrate.curve.DiscountCurve; 
import com.opengamma.analytics.financial.model.interestrate.curve.YieldAndDiscountAddZeroSpreadCurve; 
import com.opengamma.analytics.financial.model.interestrate.curve.YieldAndDiscountCurve; 
import com.opengamma.analytics.financial.model.interestrate.curve.YieldCurve; 
import com.opengamma.analytics.financial.schedule.ScheduleCalculator; 
import com.opengamma.analytics.math.curve.DoublesCurveNelsonSiegel; 
import com.opengamma.analytics.math.curve.InterpolatedDoublesCurve; 
import com.opengamma.analytics.math.interpolation.CombinedInterpolatorExtrapolatorFactory; 
import com.opengamma.analytics.math.interpolation.Interpolator1D; 
import com.opengamma.analytics.math.interpolation.Interpolator1DFactory; 
import com.opengamma.analytics.util.time.TimeCalculator; 
import com.opengamma.financial.convention.businessday.BusinessDayConvention; 
import com.opengamma.financial.convention.businessday.BusinessDayConventionFactory; 
import com.opengamma.financial.convention.calendar.Calendar; 
import com.opengamma.financial.convention.calendar.MondayToFridayCalendar; 
import com.opengamma.financial.convention.daycount.DayCount; 
import com.opengamma.financial.convention.daycount.DayCountFactory; 
import com.opengamma.util.money.Currency; 
import com.opengamma.util.money.MultipleCurrencyAmount; 
import com.opengamma.util.time.DateUtils; 
import com.opengamma.util.timeseries.DoubleTimeSeries; 
import com.opengamma.util.timeseries.zoneddatetime.ArrayZonedDateTimeDoubleTimeSeries; 
import com.opengamma.lambdava.tuple.ObjectsPair; 
import com.opengamma.lambdava.tuple.Pair; 

/**
 * Build of curve in several blocks with relevant Jacobian matrices.
 * Curves are using different types of interpolations (on cc rate, periodic rates and discount factors) and spread curves.
 */
public  class  CurveConstructionSpreadTest {
	

  //TODO: add test on futures.
  

	
  

	
  //  private static final Interpolator1D INTERPOLATOR_CS = CombinedInterpolatorExtrapolatorFactory.getInterpolator(Interpolator1DFactory.NATURAL_CUBIC_SPLINE, Interpolator1DFactory.FLAT_EXTRAPOLATOR,
  //      Interpolator1DFactory.FLAT_EXTRAPOLATOR);
  

	 // Log-linear on the discount factor = step on the instantaneous rates
  //  private static final MatrixAlgebra MATRIX_ALGEBRA = new CommonsMatrixAlgebra();

  

	
  

	
  

	
  

	
  

	
  

	
  

	
  

	
  

	
  

	
  

	
  

	
  

	
  

	
  

	
  

	
  

	
  

	

  

	

  

	
  

	
  

	
  

	
  

	

  

	
  

	

  

	
  

	

  

	
  

	
  

	
  

	

  

	
  

	
  

	
  

	

  /** Market values for the dsc USD curve */
  

	
  /** Generators for the dsc USD curve */
  

	
  /** Tenors for the dsc USD curve */
  

	

  /** Market values for the Fwd 3M USD curve */
  

	
  /** Generators for the Fwd 3M USD curve */
  

	
  /** Tenors for the Fwd 3M USD curve */
  

	

  /** Market values for the Fwd 3M USD curve */
  

	
  /** Generators for the Fwd 3M USD curve */
  

	
  /** Tenors for the Fwd 3M USD curve */
  

	

  /** Market values for the Fwd 3M USD curve */
  

	
  /** Generators for the Fwd 3M USD curve */
  

	
  /** Tenors for the Fwd 3M USD curve */
  

	

  /** Market values for the Fwd 3M USD curve */
  

	
  /** Generators for the Fwd 3M USD curve */
  

	
  /** Tenors for the Fwd 3M USD curve */
  

	

  /** Market values for the Fwd 6M USD curve */
  

	
  /** Generators for the Fwd 6M USD curve */
  

	
  /** Tenors for the Fwd 6M USD curve */
  

	

  /** Standard USD discounting curve instrument definitions */
  

	
  /** Standard USD Forward 3M curve instrument definitions */
  

	
  /** Standard USD Forward 3M curve instrument definitions */
  

	
  /** Standard USD Forward 3M curve instrument definitions */
  

	
  /** Standard USD Forward 3M curve instrument definitions */
  

	
  /** Standard USD Forward 6M curve instrument definitions */
  

	
  /** Units of curves */
  

	
  

	
  

	
  

	
  

	
  

	

  // Present Value
  

	
  

	
  

	
  

	
  

	
  // Par spread market quote
  

	
  

	

  

	
  

	
  

	
  

	

  // Instrument used for sensitivity tests
  

	
  

	
  

	
  

	
  

	

  

	
  

	

  

	

  

	

  

	

  

	

  

	

  

	

  

	

  

	

  

	

  

	

  

	

  

	

  

	

  

	

  

	

  

	

  

	

  

	

  

	

  

	
  static {
    CCY_MAP.put(CURVE_NAME_DSC_USD, CCY_USD);
    CCY_MAP.put(CURVE_NAME_FWD3_USD, CCY_USD);
    CCY_MAP.put(CURVE_NAME_FWD6_USD, CCY_USD);
  }

	

  static {
    DEFINITIONS_DSC_USD = getDefinitions(DSC_USD_MARKET_QUOTES, DSC_USD_GENERATORS, DSC_USD_TENOR, new Double[DSC_USD_MARKET_QUOTES.length]);
    DEFINITIONS_FWD3_USD = getDefinitions(FWD3_USD_MARKET_QUOTES, FWD3_USD_GENERATORS, FWD3_USD_TENOR, new Double[FWD3_USD_MARKET_QUOTES.length]);
    DEFINITIONS_FWD3_USD_2 = getDefinitions(FWD3_USD_MARKET_QUOTES_2, FWD3_USD_GENERATORS_2, FWD3_USD_TENOR_2, new Double[FWD3_USD_MARKET_QUOTES_2.length]);
    DEFINITIONS_FWD3_USD_3 = getDefinitions(FWD3_USD_MARKET_QUOTES_3, FWD3_USD_GENERATORS_3, FWD3_USD_TENOR_3, new Double[FWD3_USD_MARKET_QUOTES_3.length]);
    DEFINITIONS_FWD3_USD_4 = getDefinitions(FWD3_USD_MARKET_QUOTES_4, FWD3_USD_GENERATORS_4, FWD3_USD_TENOR_4, new Double[FWD3_USD_MARKET_QUOTES_4.length]);
    DEFINITIONS_FWD6_USD = getDefinitions(FWD6_USD_MARKET_QUOTES, FWD6_USD_GENERATORS, FWD6_USD_TENOR, new Double[FWD6_USD_MARKET_QUOTES.length]);
    for (int loopblock = 0; loopblock < NB_BLOCKS; loopblock++) {
      DEFINITIONS_UNITS[loopblock] = new InstrumentDefinition<?>[NB_UNITS[loopblock]][][];
      GENERATORS_UNITS[loopblock] = new GeneratorYDCurve[NB_UNITS[loopblock]][];
      NAMES_UNITS[loopblock] = new String[NB_UNITS[loopblock]][];
    }
    DEFINITIONS_UNITS[0][0] = new InstrumentDefinition<?>[][] {DEFINITIONS_DSC_USD };
    DEFINITIONS_UNITS[0][1] = new InstrumentDefinition<?>[][] {DEFINITIONS_FWD3_USD_2 };
    DEFINITIONS_UNITS[1][0] = new InstrumentDefinition<?>[][] {DEFINITIONS_DSC_USD };
    DEFINITIONS_UNITS[1][1] = new InstrumentDefinition<?>[][] {DEFINITIONS_FWD3_USD_2 };
    DEFINITIONS_UNITS[2][0] = new InstrumentDefinition<?>[][] {DEFINITIONS_DSC_USD };
    DEFINITIONS_UNITS[2][1] = new InstrumentDefinition<?>[][] {DEFINITIONS_FWD3_USD_2 };
    DEFINITIONS_UNITS[3][0] = new InstrumentDefinition<?>[][] {DEFINITIONS_DSC_USD };
    DEFINITIONS_UNITS[3][1] = new InstrumentDefinition<?>[][] {DEFINITIONS_FWD3_USD_2 };
    DEFINITIONS_UNITS[3][2] = new InstrumentDefinition<?>[][] {DEFINITIONS_FWD6_USD };
    DEFINITIONS_UNITS[4][0] = new InstrumentDefinition<?>[][] {DEFINITIONS_DSC_USD };
    DEFINITIONS_UNITS[4][1] = new InstrumentDefinition<?>[][] {DEFINITIONS_FWD3_USD_2 };
    DEFINITIONS_UNITS[4][2] = new InstrumentDefinition<?>[][] {DEFINITIONS_FWD6_USD };
    DEFINITIONS_UNITS[5][0] = new InstrumentDefinition<?>[][] {DEFINITIONS_FWD3_USD_3, DEFINITIONS_DSC_USD };
    DEFINITIONS_UNITS[6][0] = new InstrumentDefinition<?>[][] {DEFINITIONS_FWD3_USD_3, DEFINITIONS_DSC_USD };
    DEFINITIONS_UNITS[7][0] = new InstrumentDefinition<?>[][] {DEFINITIONS_DSC_USD };
    DEFINITIONS_UNITS[7][1] = new InstrumentDefinition<?>[][] {DEFINITIONS_FWD3_USD_4 };
    DEFINITIONS_UNITS[8][0] = new InstrumentDefinition<?>[][] {DEFINITIONS_DSC_USD };
    DEFINITIONS_UNITS[9][0] = new InstrumentDefinition<?>[][] {DEFINITIONS_DSC_USD };
    final GeneratorYDCurve genIntDQ = new GeneratorCurveYieldInterpolated(MATURITY_CALCULATOR, INTERPOLATOR_DQ);
    final GeneratorYDCurve genIntLin = new GeneratorCurveYieldInterpolated(MATURITY_CALCULATOR, INTERPOLATOR_LINEAR);
    final int compoundingRate = 1;
    final GeneratorYDCurve genIntRPLin = new GeneratorCurveYieldPeriodicInterpolated(MATURITY_CALCULATOR, compoundingRate, INTERPOLATOR_LINEAR);
    final GeneratorYDCurve genIntDFLL = new GeneratorCurveDiscountFactorInterpolated(MATURITY_CALCULATOR, INTERPOLATOR_LL);
    final GeneratorYDCurve genNS = new GeneratorCurveYieldNelsonSiegel();
    final GeneratorYDCurve genInt0 = new GeneratorCurveYieldInterpolatedAnchor(MATURITY_CALCULATOR, INTERPOLATOR_LINEAR);
    final GeneratorYDCurve genAddExistFwd3 = new GeneratorCurveAddYieldExisiting(genIntDQ, false, CURVE_NAME_FWD3_USD);
    final LocalDate startTOY = LocalDate.of(2011, 12, 30);
    final LocalDate endTOY = LocalDate.of(2012, 1, 2);
    final double spreadTOY = 0.0025; // 25bps
    final double dfTOY = 1.0 / (1 + DAY_COUNT_CASH.getDayCountFraction(startTOY, endTOY) * spreadTOY);
    final LocalDate startTOQ = LocalDate.of(2012, 3, 30);
    final LocalDate endTOQ = LocalDate.of(2012, 4, 2);
    final double spreadTOQ = 0.0010; // 25bps
    final double dfTOQ = 1.0 / (1 + DAY_COUNT_CASH.getDayCountFraction(startTOQ, endTOQ) * spreadTOQ);
    final double[] times = {TimeCalculator.getTimeBetween(NOW, startTOY), TimeCalculator.getTimeBetween(NOW, endTOY), TimeCalculator.getTimeBetween(NOW, startTOQ),
        TimeCalculator.getTimeBetween(NOW, endTOQ) };
    final double[] df = {1.0, dfTOY, dfTOY, dfTOY * dfTOQ };
    final YieldAndDiscountCurve curveTOY = new DiscountCurve("TOY", new InterpolatedDoublesCurve(times, df, INTERPOLATOR_LINEAR, true));
    final GeneratorYDCurve genAddFixed = new GeneratorCurveAddYieldFixed(genIntDQ, false, curveTOY);
    final GeneratorYDCurve genIntDQ0 = new GeneratorCurveYieldInterpolatedAnchor(MATURITY_CALCULATOR, INTERPOLATOR_DQ);
    final int[] nbParameters = {5, DSC_USD_MARKET_QUOTES.length - 5 };
    final GeneratorYDCurve gen2Blocks = new GeneratorCurveAddYieldNb(new GeneratorYDCurve[] {genIntDFLL, genIntDQ0 }, nbParameters, false);
    GENERATORS_UNITS[0][0] = new GeneratorYDCurve[] {genIntLin };
    GENERATORS_UNITS[0][1] = new GeneratorYDCurve[] {genIntLin };
    GENERATORS_UNITS[1][0] = new GeneratorYDCurve[] {genIntRPLin };
    GENERATORS_UNITS[1][1] = new GeneratorYDCurve[] {genIntRPLin };
    GENERATORS_UNITS[2][0] = new GeneratorYDCurve[] {genIntDFLL };
    GENERATORS_UNITS[2][1] = new GeneratorYDCurve[] {genIntDFLL };
    // 3xinterpolated / 2xinterpolated + spread over existing
    GENERATORS_UNITS[3][0] = new GeneratorYDCurve[] {genIntDQ };
    GENERATORS_UNITS[3][1] = new GeneratorYDCurve[] {genIntDQ };
    GENERATORS_UNITS[3][2] = new GeneratorYDCurve[] {genIntDQ };
    GENERATORS_UNITS[4][0] = new GeneratorYDCurve[] {genIntDQ };
    GENERATORS_UNITS[4][1] = new GeneratorYDCurve[] {genIntDQ };
    GENERATORS_UNITS[4][2] = new GeneratorYDCurve[] {genAddExistFwd3 };
    // 2xinterpolated / interpolated + spread over existing
    GENERATORS_UNITS[5][0] = new GeneratorYDCurve[] {genIntDQ, genIntDQ };
    GENERATORS_UNITS[6][0] = new GeneratorYDCurve[] {genIntDQ, genAddExistFwd3 };
    // interpolated + functional+interpolated
    GENERATORS_UNITS[7][0] = new GeneratorYDCurve[] {genIntDQ };
    GENERATORS_UNITS[7][1] = new GeneratorYDCurve[] {new GeneratorCurveAddYield(new GeneratorYDCurve[] {genNS, genInt0 }, false) };
    GENERATORS_UNITS[8][0] = new GeneratorYDCurve[] {genAddFixed };
    GENERATORS_UNITS[9][0] = new GeneratorYDCurve[] {gen2Blocks };
    NAMES_UNITS[0][0] = new String[] {CURVE_NAME_DSC_USD };
    NAMES_UNITS[0][1] = new String[] {CURVE_NAME_FWD3_USD };
    NAMES_UNITS[1][0] = new String[] {CURVE_NAME_DSC_USD };
    NAMES_UNITS[1][1] = new String[] {CURVE_NAME_FWD3_USD };
    NAMES_UNITS[2][0] = new String[] {CURVE_NAME_DSC_USD };
    NAMES_UNITS[2][1] = new String[] {CURVE_NAME_FWD3_USD };
    NAMES_UNITS[3][0] = new String[] {CURVE_NAME_DSC_USD };
    NAMES_UNITS[3][1] = new String[] {CURVE_NAME_FWD3_USD };
    NAMES_UNITS[3][2] = new String[] {CURVE_NAME_FWD6_USD };
    NAMES_UNITS[4][0] = new String[] {CURVE_NAME_DSC_USD };
    NAMES_UNITS[4][1] = new String[] {CURVE_NAME_FWD3_USD };
    NAMES_UNITS[4][2] = new String[] {CURVE_NAME_FWD6_USD };
    NAMES_UNITS[5][0] = new String[] {CURVE_NAME_FWD3_USD, CURVE_NAME_DSC_USD };
    NAMES_UNITS[6][0] = new String[] {CURVE_NAME_FWD3_USD, CURVE_NAME_DSC_USD };
    NAMES_UNITS[7][0] = new String[] {CURVE_NAME_DSC_USD };
    NAMES_UNITS[7][1] = new String[] {CURVE_NAME_FWD3_USD };
    NAMES_UNITS[8][0] = new String[] {CURVE_NAME_DSC_USD };
    NAMES_UNITS[9][0] = new String[] {CURVE_NAME_DSC_USD };
  }


}
