/**
 * Copyright (C) 2012 - present by OpenGamma Inc. and the OpenGamma group of companies
 * 
 * Please see distribution for license.
 */
package com.opengamma.analytics.financial.curve; 

import static org.testng.AssertJUnit.assertEquals; 
import static org.threeten.bp.temporal.ChronoUnit.DAYS; 
import static org.threeten.bp.temporal.ChronoUnit.MONTHS; 
import static org.threeten.bp.temporal.ChronoUnit.YEARS; 

import java.util.ArrayList; 
import java.util.Arrays; 
import java.util.HashMap; 
import java.util.HashSet; 
import java.util.LinkedHashMap; 
import java.util.List; 

import org.apache.commons.lang.ArrayUtils; 
import org.testng.annotations.BeforeSuite; 
import org.testng.annotations.Test; 
import org.threeten.bp.Period; 
import org.threeten.bp.ZonedDateTime; 

import com.opengamma.analytics.financial.calculator.MarketQuoteSensitivityBlockCalculator; 
import com.opengamma.analytics.financial.calculator.PresentValueConvertedCalculator; 
import com.opengamma.analytics.financial.calculator.PresentValueCurveSensitivityConvertedCalculator; 
import com.opengamma.analytics.financial.calculator.PresentValueCurveSensitivityMCSCalculator; 
import com.opengamma.analytics.financial.calculator.PresentValueMCACalculator; 
import com.opengamma.analytics.financial.curve.building.CurveBuildingBlock; 
import com.opengamma.analytics.financial.curve.building.CurveBuildingBlockBundle; 
import com.opengamma.analytics.financial.curve.building.MultipleYieldCurveFinderGeneratorDataBundle; 
import com.opengamma.analytics.financial.curve.building.MultipleYieldCurveFinderGeneratorFunction; 
import com.opengamma.analytics.financial.curve.building.MultipleYieldCurveFinderGeneratorJacobian; 
import com.opengamma.analytics.financial.curve.generator.GeneratorCurveYieldInterpolated; 
import com.opengamma.analytics.financial.curve.generator.GeneratorYDCurve; 
import com.opengamma.analytics.financial.curve.sensitivity.ParameterSensitivity; 
import com.opengamma.analytics.financial.curve.sensitivity.ParameterSensitivityBlockCalculator; 
import com.opengamma.analytics.financial.curve.sensitivity.ParameterSensitivityCalculator; 
import com.opengamma.analytics.financial.forex.definition.ForexSwapDefinition; 
import com.opengamma.analytics.financial.forex.method.FXMatrix; 
import com.opengamma.analytics.financial.instrument.InstrumentDefinition; 
import com.opengamma.analytics.financial.instrument.cash.CashDefinition; 
import com.opengamma.analytics.financial.instrument.fra.ForwardRateAgreementDefinition; 
import com.opengamma.analytics.financial.instrument.index.GeneratorDeposit; 
import com.opengamma.analytics.financial.instrument.index.GeneratorDepositON; 
import com.opengamma.analytics.financial.instrument.index.GeneratorForexSwap; 
import com.opengamma.analytics.financial.instrument.index.GeneratorInstrument; 
import com.opengamma.analytics.financial.instrument.index.GeneratorSwapFixedIbor; 
import com.opengamma.analytics.financial.instrument.index.GeneratorSwapFixedIborMaster; 
import com.opengamma.analytics.financial.instrument.index.GeneratorSwapFixedON; 
import com.opengamma.analytics.financial.instrument.index.GeneratorSwapIborIbor; 
import com.opengamma.analytics.financial.instrument.index.GeneratorSwapXCcyIborIbor; 
import com.opengamma.analytics.financial.instrument.index.IborIndex; 
import com.opengamma.analytics.financial.instrument.index.IndexIborMaster; 
import com.opengamma.analytics.financial.instrument.index.IndexON; 
import com.opengamma.analytics.financial.instrument.swap.SwapFixedIborDefinition; 
import com.opengamma.analytics.financial.instrument.swap.SwapFixedONDefinition; 
import com.opengamma.analytics.financial.instrument.swap.SwapIborIborDefinition; 
import com.opengamma.analytics.financial.instrument.swap.SwapXCcyIborIborDefinition; 
import com.opengamma.analytics.financial.interestrate.InstrumentDerivative; 
import com.opengamma.analytics.financial.interestrate.InstrumentDerivativeVisitor; 
import com.opengamma.analytics.financial.interestrate.InterestRateCurveSensitivity; 
import com.opengamma.analytics.financial.interestrate.LastTimeCalculator; 
import com.opengamma.analytics.financial.interestrate.ParSpreadMarketQuoteCalculator; 
import com.opengamma.analytics.financial.interestrate.ParSpreadMarketQuoteCurveSensitivityCalculator; 
import com.opengamma.analytics.financial.interestrate.YieldCurveBundle; 
import com.opengamma.analytics.financial.interestrate.payments.derivative.Payment; 
import com.opengamma.analytics.financial.interestrate.swap.derivative.Swap; 
import com.opengamma.analytics.financial.schedule.ScheduleCalculator; 
import com.opengamma.analytics.math.function.Function1D; 
import com.opengamma.analytics.math.interpolation.CombinedInterpolatorExtrapolatorFactory; 
import com.opengamma.analytics.math.interpolation.Interpolator1D; 
import com.opengamma.analytics.math.interpolation.Interpolator1DFactory; 
import com.opengamma.analytics.math.linearalgebra.DecompositionFactory; 
import com.opengamma.analytics.math.matrix.CommonsMatrixAlgebra; 
import com.opengamma.analytics.math.matrix.DoubleMatrix1D; 
import com.opengamma.analytics.math.matrix.DoubleMatrix2D; 
import com.opengamma.analytics.math.matrix.MatrixAlgebra; 
import com.opengamma.analytics.math.rootfinding.newton.BroydenVectorRootFinder; 
import com.opengamma.financial.convention.businessday.BusinessDayConvention; 
import com.opengamma.financial.convention.businessday.BusinessDayConventionFactory; 
import com.opengamma.financial.convention.calendar.Calendar; 
import com.opengamma.financial.convention.calendar.MondayToFridayCalendar; 
import com.opengamma.financial.convention.daycount.DayCount; 
import com.opengamma.financial.convention.daycount.DayCountFactory; 
import com.opengamma.util.money.Currency; 
import com.opengamma.util.time.DateUtils; 
import com.opengamma.util.timeseries.DoubleTimeSeries; 
import com.opengamma.util.timeseries.zoneddatetime.ArrayZonedDateTimeDoubleTimeSeries; 
import com.opengamma.lambdava.tuple.ObjectsPair; 
import com.opengamma.lambdava.tuple.Pair; 

/**
 * Build of curve in several currencies in several blocks with relevant Jacobian matrices.
 * Currencies: USD (2 curves), EUR (2 curves), JPY (3 curves)
 */
public  class  CurveConstructionXCcyTest {
	

  //  private static final Interpolator1D INTERPOLATOR = CombinedInterpolatorExtrapolatorFactory.getInterpolator(Interpolator1DFactory.DOUBLE_QUADRATIC,
  //      Interpolator1DFactory.LINEAR_EXTRAPOLATOR, Interpolator1DFactory.FLAT_EXTRAPOLATOR);
  

	
  

	

  

	
  

	
  

	
  

	
  

	
  

	
  

	
  

	
  

	
  

	
  

	
  

	
  

	
  

	
  

	
  

	
  

	
  

	
  

	
  

	
  

	
  

	
  

	
  

	
  

	
  

	
  

	
  

	
  

	
  

	
  

	
  

	
  

	
  

	 // Spread on EUR leg
  

	 // Spread on JPY leg
  

	 // Spread on JPY leg
  

	

  

	

  

	
  

	
  

	
  

	
  

	

  

	
  

	

  

	
  

	

  

	
  

	
  

	
  

	
  

	
  

	
  

	
  

	
  

	
  

	
  

	
  

	

  

	
  

	
  

	
  

	
  

	
  

	
  

	
  

	

  /** Market values for the dsc USD curve */
  

	
  /** Generators for the dsc USD curve */
  

	
  /** Tenors for the dsc USD curve */
  

	

  /** Market values for the Fwd 3M USD curve */
  

	
  /** Generators for the Fwd 3M USD curve */
  

	
  /** Tenors for the Fwd 3M USD curve */
  

	

  /** Market values for the dsc EUR curve */
  

	
  //  public static final double[] DSC_EUR_MARKET_QUOTES = new double[] {0.0010, 0.0010, 0.0004 * FX_EURUSD, 0.0009 * FX_EURUSD, 0.0015 * FX_EURUSD, 0.0035 * FX_EURUSD, 0.0050 * FX_EURUSD,
  //    0.0060 * FX_EURUSD, -0.0050, -0.0050, -0.0050, -0.0045, -0.0040};
  /** Generators for the dsc EUR curve */
  

	
  /** Tenors for the dsc EUR curve */
  

	
  //  /** FX rqtes for the dsc EUR curve */
  //  public static final Double[] DSC_EUR_FX_RATE = new Double[] {FX_EURUSD, FX_EURUSD, FX_EURUSD, FX_EURUSD, FX_EURUSD, FX_EURUSD, FX_EURUSD, FX_EURUSD, FX_EURUSD, FX_EURUSD, FX_EURUSD, FX_EURUSD,
  //      FX_EURUSD};

  /** Market values for the Fwd 3M EUR curve */
  

	
  /** Generators for the Fwd 3M USD curve */
  

	
  /** Tenors for the Fwd 3M USD curve */
  

	

  /** Market values for the dsc JPY curve */
  

	
  /** Generators for the dsc EUR curve */
  

	
  /** Tenors for the dsc EUR curve */
  

	
  //  /** FX rqtes for the dsc EUR curve */
  //  public static final Double[] DSC_JPY_FX_RATE = new Double[] {FX_USDJPY, FX_USDJPY, FX_USDJPY, FX_USDJPY, FX_USDJPY, FX_USDJPY, FX_USDJPY, FX_USDJPY, FX_USDJPY, FX_USDJPY, FX_USDJPY, FX_USDJPY,
  //      FX_USDJPY};

  /** Market values for the Fwd 3M JPY curve */
  

	
  /** Generators for the Fwd 3M JPY curve */
  

	
  /** Tenors for the Fwd 3M JPY curve */
  

	

  /** Market values for the Fwd 6M JPY curve */
  

	
  /** Generators for the Fwd 6M JPY curve */
  

	
  /** Tenors for the Fwd 6M JPY curve */
  

	

  /** Standard USD discounting curve instrument definitions */
  

	
  /** Standard USD Forward 3M curve instrument definitions */
  

	
  /** Standard EUR discounting curve instrument definitions */
  

	
  /** Standard EUR Forward 3M curve instrument definitions */
  

	
  /** Standard JPY discounting curve instrument definitions */
  

	
  /** Standard JPY Forward 3M curve instrument definitions */
  

	
  /** Standard JPY Forward 6M curve instrument definitions */
  

	
  /** Units of curves */
  

	
  

	
  

	
  

	
  

	
  

	

  // Present Value
  

	
  

	
  

	
  

	
  

	
  // Par spread market quote
  

	
  

	

  

	
  

	
  

	
  

	
  

	
  

	
  

	

  

	

  

	

  

	

  

	

  

	

  

	

  

	

  

	

  

	

  

	

  

	

  

	

  

	

  

	

  

	

  

	

  

	

  

	

  

	

  

	

  

	

  

	

  

	

  

	
  static {
    CCY_MAP.put(CURVE_NAME_DSC_USD, CCY_USD);
    CCY_MAP.put(CURVE_NAME_FWD3_USD, CCY_USD);
    CCY_MAP.put(CURVE_NAME_DSC_EUR, CCY_EUR);
    CCY_MAP.put(CURVE_NAME_FWD3_EUR, CCY_EUR);
    CCY_MAP.put(CURVE_NAME_DSC_JPY, CCY_JPY);
    CCY_MAP.put(CURVE_NAME_FWD3_JPY, CCY_JPY);
    CCY_MAP.put(CURVE_NAME_FWD6_JPY, CCY_JPY);
    FX_MATRIX.addCurrency(CCY_EUR, CCY_USD, FX_EURUSD);
    FX_MATRIX.addCurrency(CCY_JPY, CCY_USD, 1 / FX_USDJPY);
  }

	

  static {
    DEFINITIONS_DSC_USD = getDefinitions(DSC_1_MARKET_QUOTES, DSC_1_GENERATORS, DSC_1_TENOR);
    DEFINITIONS_FWD3_USD = getDefinitions(FWD_1_MARKET_QUOTES, FWD_1_GENERATORS, FWD_1_TENOR);
    DEFINITIONS_DSC_EUR = getDefinitions(DSC_EUR_MARKET_QUOTES, DSC_EUR_GENERATORS, DSC_EUR_TENOR);
    DEFINITIONS_FWD3_EUR = getDefinitions(FWD_EUR_MARKET_QUOTES, FWD_EUR_GENERATORS, FWD_EUR_TENOR);
    DEFINITIONS_DSC_JPY = getDefinitions(DSC_JPY_MARKET_QUOTES, DSC_JPY_GENERATORS, DSC_JPY_TENOR);
    DEFINITIONS_FWD3_JPY = getDefinitions(FWD3_JPY_MARKET_QUOTES, FWD3_JPY_GENERATORS, FWD3_JPY_TENOR);
    DEFINITIONS_FWD6_JPY = getDefinitions(FWD6_JPY_MARKET_QUOTES, FWD6_JPY_GENERATORS, FWD6_JPY_TENOR);
    DEFINITIONS_UNITS[0] = new InstrumentDefinition<?>[NB_UNITS[0]][][];
    DEFINITIONS_UNITS[1] = new InstrumentDefinition<?>[NB_UNITS[1]][][];
    DEFINITIONS_UNITS[2] = new InstrumentDefinition<?>[NB_UNITS[2]][][];
    DEFINITIONS_UNITS[0][0] = new InstrumentDefinition<?>[][] {DEFINITIONS_DSC_USD };
    DEFINITIONS_UNITS[0][1] = new InstrumentDefinition<?>[][] {DEFINITIONS_FWD3_USD };
    DEFINITIONS_UNITS[0][2] = new InstrumentDefinition<?>[][] {DEFINITIONS_DSC_EUR, DEFINITIONS_FWD3_EUR };
    DEFINITIONS_UNITS[1][0] = new InstrumentDefinition<?>[][] {DEFINITIONS_DSC_USD };
    DEFINITIONS_UNITS[1][1] = new InstrumentDefinition<?>[][] {DEFINITIONS_FWD3_USD };
    DEFINITIONS_UNITS[1][2] = new InstrumentDefinition<?>[][] {DEFINITIONS_DSC_JPY, DEFINITIONS_FWD3_JPY, DEFINITIONS_FWD6_JPY };
    DEFINITIONS_UNITS[2][0] = new InstrumentDefinition<?>[][] {DEFINITIONS_DSC_USD, DEFINITIONS_FWD3_USD, DEFINITIONS_DSC_JPY, DEFINITIONS_FWD3_JPY, DEFINITIONS_FWD6_JPY };
    final GeneratorYDCurve genInt = new GeneratorCurveYieldInterpolated(MATURITY_CALCULATOR, INTERPOLATOR);
    GENERATORS_UNITS[0] = new GeneratorYDCurve[NB_UNITS[0]][];
    GENERATORS_UNITS[1] = new GeneratorYDCurve[NB_UNITS[1]][];
    GENERATORS_UNITS[2] = new GeneratorYDCurve[NB_UNITS[2]][];
    GENERATORS_UNITS[0][0] = new GeneratorYDCurve[] {genInt };
    GENERATORS_UNITS[0][1] = new GeneratorYDCurve[] {genInt };
    GENERATORS_UNITS[0][2] = new GeneratorYDCurve[] {genInt, genInt };
    GENERATORS_UNITS[1][0] = new GeneratorYDCurve[] {genInt };
    GENERATORS_UNITS[1][1] = new GeneratorYDCurve[] {genInt };
    GENERATORS_UNITS[1][2] = new GeneratorYDCurve[] {genInt, genInt, genInt };
    GENERATORS_UNITS[2][0] = new GeneratorYDCurve[] {genInt, genInt, genInt, genInt, genInt };
    NAMES_UNITS[0] = new String[NB_UNITS[0]][];
    NAMES_UNITS[1] = new String[NB_UNITS[1]][];
    NAMES_UNITS[2] = new String[NB_UNITS[2]][];
    NAMES_UNITS[0][0] = new String[] {CURVE_NAME_DSC_USD };
    NAMES_UNITS[0][1] = new String[] {CURVE_NAME_FWD3_USD };
    NAMES_UNITS[0][2] = new String[] {CURVE_NAME_DSC_EUR, CURVE_NAME_FWD3_EUR };
    NAMES_UNITS[1][0] = new String[] {CURVE_NAME_DSC_USD };
    NAMES_UNITS[1][1] = new String[] {CURVE_NAME_FWD3_USD };
    NAMES_UNITS[1][2] = new String[] {CURVE_NAME_DSC_JPY, CURVE_NAME_FWD3_JPY, CURVE_NAME_FWD6_JPY };
    NAMES_UNITS[2][0] = new String[] {CURVE_NAME_DSC_USD, CURVE_NAME_FWD3_USD, CURVE_NAME_DSC_JPY, CURVE_NAME_FWD3_JPY, CURVE_NAME_FWD6_JPY };
  }


}
