/**
 * Copyright (C) 2009 - present by OpenGamma Inc. and the OpenGamma group of companies
 * 
 * Please see distribution for license.
 */
package com.opengamma.analytics.math.rootfinding; 

import static com.opengamma.analytics.math.interpolation.Interpolator1DFactory.FLAT_EXTRAPOLATOR; 
import static com.opengamma.analytics.math.interpolation.Interpolator1DFactory.LINEAR_EXTRAPOLATOR; 
import static org.testng.AssertJUnit.assertEquals; 
import static org.threeten.bp.temporal.ChronoUnit.MONTHS; 

import java.util.ArrayList; 
import java.util.Arrays; 
import java.util.LinkedHashMap; 
import java.util.List; 
import java.util.Map; 

import org.slf4j.Logger; 
import org.slf4j.LoggerFactory; 
import org.testng.annotations.Test; 
import org.threeten.bp.Period; 

import cern.jet.random.engine.MersenneTwister; 
import cern.jet.random.engine.MersenneTwister64; 
import cern.jet.random.engine.RandomEngine; 

import com.opengamma.analytics.financial.forex.method.FXMatrix; 
import com.opengamma.analytics.financial.instrument.index.IborIndex; 
import com.opengamma.analytics.financial.interestrate.InstrumentDerivative; 
import com.opengamma.analytics.financial.interestrate.InstrumentDerivativeVisitor; 
import com.opengamma.analytics.financial.interestrate.MultipleYieldCurveFinderDataBundle; 
import com.opengamma.analytics.financial.interestrate.MultipleYieldCurveFinderFunction; 
import com.opengamma.analytics.financial.interestrate.MultipleYieldCurveFinderJacobian; 
import com.opengamma.analytics.financial.interestrate.ParRateCalculator; 
import com.opengamma.analytics.financial.interestrate.ParRateCurveSensitivityCalculator; 
import com.opengamma.analytics.financial.interestrate.RateReplacingInterestRateDerivativeVisitor; 
import com.opengamma.analytics.financial.interestrate.YieldCurveBundle; 
import com.opengamma.analytics.financial.interestrate.annuity.derivative.AnnuityCouponFixed; 
import com.opengamma.analytics.financial.interestrate.annuity.derivative.AnnuityCouponIbor; 
import com.opengamma.analytics.financial.interestrate.swap.derivative.FixedFloatSwap; 
import com.opengamma.analytics.financial.model.interestrate.curve.YieldAndDiscountCurve; 
import com.opengamma.analytics.financial.model.interestrate.curve.YieldCurve; 
import com.opengamma.analytics.math.curve.InterpolatedDoublesCurve; 
import com.opengamma.analytics.math.differentiation.VectorFieldFirstOrderDifferentiator; 
import com.opengamma.analytics.math.function.Function1D; 
import com.opengamma.analytics.math.interpolation.CombinedInterpolatorExtrapolatorFactory; 
import com.opengamma.analytics.math.interpolation.Interpolator1D; 
import com.opengamma.analytics.math.interpolation.Interpolator1DFactory; 
import com.opengamma.analytics.math.matrix.DoubleMatrix1D; 
import com.opengamma.analytics.math.matrix.DoubleMatrix2D; 
import com.opengamma.analytics.math.rootfinding.newton.BroydenVectorRootFinder; 
import com.opengamma.analytics.math.rootfinding.newton.NewtonDefaultVectorRootFinder; 
import com.opengamma.analytics.math.rootfinding.newton.NewtonVectorRootFinder; 
import com.opengamma.analytics.math.rootfinding.newton.ShermanMorrisonVectorRootFinder; 
import com.opengamma.analytics.math.statistics.distribution.NormalDistribution; 
import com.opengamma.financial.convention.businessday.BusinessDayConvention; 
import com.opengamma.financial.convention.businessday.BusinessDayConventionFactory; 
import com.opengamma.financial.convention.calendar.Calendar; 
import com.opengamma.financial.convention.calendar.MondayToFridayCalendar; 
import com.opengamma.financial.convention.daycount.DayCount; 
import com.opengamma.financial.convention.daycount.DayCountFactory; 
import com.opengamma.util.money.Currency; 
import com.opengamma.util.monitor.OperationTimer; 
import com.opengamma.util.tuple.DoublesPair; 
import com.opengamma.lambdava.tuple.DoublesPair; 

/**
Timings on Richard's mac pro from 24/08/2010 with 200 warm ups and 1000 benchmark cycles

18:35:26.423 [main] INFO  c.o.m.r.YieldCurveBootStrapTest - 14947ms-processing 1000 cycles on default Newton, single curve
18:36:39.163 [main] INFO  c.o.m.r.YieldCurveBootStrapTest - 60618ms-processing 1000 cycles on default Newton, single curve, finite difference
18:38:47.446 [main] INFO  c.o.m.r.YieldCurveBootStrapTest - 106922ms-processing 1000 cycles on default Newton, single curve FD interpolator sensitivity
18:39:05.593 [main] INFO  c.o.m.r.YieldCurveBootStrapTest - 15075ms-processing 1000 cycles on default Newton, double curve
18:40:23.854 [main] INFO  c.o.m.r.YieldCurveBootStrapTest - 65177ms-processing 1000 cycles on default Newton, double curve, finite difference
18:41:42.041 [main] INFO  c.o.m.r.YieldCurveBootStrapTest - 65185ms-processing 1000 cycles on default Newton, double curve FD interpolator sensitivity
18:41:50.663 [main] INFO  c.o.m.r.YieldCurveBootStrapTest - 7173ms-processing 1000 cycles on Sherman Morrison, single curve
18:42:07.014 [main] INFO  c.o.m.r.YieldCurveBootStrapTest - 13614ms-processing 1000 cycles on Sherman Morrison, single curve, finite difference
18:42:31.229 [main] INFO  c.o.m.r.YieldCurveBootStrapTest - 20172ms-processing 1000 cycles on Sherman Morrison, single curve FD sensitivity
18:42:38.811 [main] INFO  c.o.m.r.YieldCurveBootStrapTest - 6320ms-processing 1000 cycles on Sherman Morrison, double curve
18:42:54.001 [main] INFO  c.o.m.r.YieldCurveBootStrapTest - 12657ms-processing 1000 cycles on Sherman Morrisonn, double curve, finite difference
18:44:12.220 [main] INFO  c.o.m.r.YieldCurveBootStrapTest - 65157ms-processing 1000 cycles on Sherman Morrison, double curve FD sensitivity
18:44:21.057 [main] INFO  c.o.m.r.YieldCurveBootStrapTest - 7344ms-processing 1000 cycles on Broyden, single curve
18:44:37.649 [main] INFO  c.o.m.r.YieldCurveBootStrapTest - 13829ms-processing 1000 cycles on Broyden, single curve, finite difference
18:45:02.122 [main] INFO  c.o.m.r.YieldCurveBootStrapTest - 20385ms-processing 1000 cycles on Broyden, single curve, FD sensitivity
18:45:09.954 [main] INFO  c.o.m.r.YieldCurveBootStrapTest - 6522ms-processing 1000 cycles on Broyden, double curve
18:45:25.414 [main] INFO  c.o.m.r.YieldCurveBootStrapTest - 12884ms-processing 1000 cycles on Broyden, double curve, finite difference
18:46:44.129 [main] INFO  c.o.m.r.YieldCurveBootStrapTest - 65560ms-processing 1000 cycles on Broyden, double curve FD sensitivity
 */

    YieldCurveBootStrapTest {
	
  

	
  

	
  

	
  

	
  

	

  

	
  

	
  

	
  

	

  

	
  

	
  

	
  

	
  

	
  

	
  

	

  

	
  

	
  

	
  

	
  

	
  

	
  

	
  

	
  

	
  

	
  

	
  

	
  

	
  

	
  

	

  

	
  

	
  

	
  

	
  

	
  

	
  

	
  

	

  static {
    final int[] payments = new int[] {1, 2, 3, 4, 6, 8, 10, 14, 20, 30, 40, 50, 60 };

    FUNDING_CURVE_TIMES = new double[] {1, 2, 5, 10, 20, 31 };
    LIBOR_CURVE_TIMES = new double[] {0.5, 1, 2, 5, 10, 20, 31 };
    FUNDING_YIELDS = new double[] {0.021, 0.036, 0.06, 0.054, 0.049, 0.044 };
    LIBOR_YIELDS = new double[] {0.01, 0.02, 0.035, 0.06, 0.055, 0.05, 0.045 };

    final int n = payments.length;
    TIME_GRID = new double[n];
    SWAP_VALUES = new double[n];
    SINGLE_CURVE_INSTRUMENTS = new ArrayList<InstrumentDerivative>();
    DOUBLE_CURVE_INSTRUMENTS = new ArrayList<InstrumentDerivative>();
    EXTRAPOLATOR = CombinedInterpolatorExtrapolatorFactory.getInterpolator(Interpolator1DFactory.DOUBLE_QUADRATIC, LINEAR_EXTRAPOLATOR, FLAT_EXTRAPOLATOR);

    final YieldCurveBundle curveBundle = new YieldCurveBundle();
    FUNDING_CURVE = makeYieldCurve(FUNDING_YIELDS, FUNDING_CURVE_TIMES, EXTRAPOLATOR);
    LIBOR_CURVE = makeYieldCurve(LIBOR_YIELDS, LIBOR_CURVE_TIMES, EXTRAPOLATOR);
    curveBundle.setCurve(FUNDING_CURVE_NAME, FUNDING_CURVE);
    curveBundle.setCurve(LIBOR_CURVE_NAME, LIBOR_CURVE);

    InstrumentDerivative instrument;
    double swapRate;
    for (int i = 0; i < n; i++) {
      instrument = setupSwap(payments[i], FUNDING_CURVE_NAME, LIBOR_CURVE_NAME);
      swapRate = instrument.accept(SWAP_RATE_CALCULATOR, curveBundle);
      instrument = setParSwapRate((FixedFloatSwap) instrument, swapRate);

      DOUBLE_CURVE_INSTRUMENTS.add(instrument);
      instrument = setupSwap(payments[i], LIBOR_CURVE_NAME, LIBOR_CURVE_NAME);
      instrument = setParSwapRate((FixedFloatSwap) instrument, swapRate);
      SINGLE_CURVE_INSTRUMENTS.add(instrument);
      TIME_GRID[i] = payments[i] * 0.5;
      SWAP_VALUES[i] = swapRate;
    }

    LinkedHashMap<String, Interpolator1D> unknownCurveInterpolators = new LinkedHashMap<String, Interpolator1D>();
    LinkedHashMap<String, double[]> unknownCurveNodes = new LinkedHashMap<String, double[]>();
    unknownCurveInterpolators.put(LIBOR_CURVE_NAME, EXTRAPOLATOR);
    unknownCurveNodes.put(LIBOR_CURVE_NAME, TIME_GRID);
    MultipleYieldCurveFinderDataBundle data = new MultipleYieldCurveFinderDataBundle(SINGLE_CURVE_INSTRUMENTS, SWAP_VALUES, null, unknownCurveNodes,
        unknownCurveInterpolators, false, FX_MATRIX);
    SINGLE_CURVE_FINDER = new MultipleYieldCurveFinderFunction(data, PAR_RATE_CALCULATOR);
    SINGLE_CURVE_JACOBIAN = new MultipleYieldCurveFinderJacobian(data, PAR_RATE_SENSITIVITY_CALCULATOR);

    data = new MultipleYieldCurveFinderDataBundle(SINGLE_CURVE_INSTRUMENTS, new double[SINGLE_CURVE_INSTRUMENTS.size()], null, unknownCurveNodes,
        unknownCurveInterpolators, true, FX_MATRIX);
    SINGLE_CURVE_JACOBIAN_WITH_FD_INTERPOLATOR_SENSITIVITY = new MultipleYieldCurveFinderJacobian(data, PAR_RATE_SENSITIVITY_CALCULATOR);

    unknownCurveInterpolators = new LinkedHashMap<String, Interpolator1D>();
    unknownCurveNodes = new LinkedHashMap<String, double[]>();
    unknownCurveInterpolators.put(FUNDING_CURVE_NAME, EXTRAPOLATOR);
    unknownCurveInterpolators.put(LIBOR_CURVE_NAME, EXTRAPOLATOR);
    unknownCurveNodes.put(FUNDING_CURVE_NAME, FUNDING_CURVE_TIMES);
    unknownCurveNodes.put(LIBOR_CURVE_NAME, LIBOR_CURVE_TIMES);
    data = new MultipleYieldCurveFinderDataBundle(DOUBLE_CURVE_INSTRUMENTS, SWAP_VALUES, null, unknownCurveNodes, unknownCurveInterpolators, false,
        FX_MATRIX);

    DOUBLE_CURVE_FINDER = new MultipleYieldCurveFinderFunction(data, PAR_RATE_CALCULATOR);
    DOUBLE_CURVE_JACOBIAN = new MultipleYieldCurveFinderJacobian(data, PAR_RATE_SENSITIVITY_CALCULATOR);

    data = new MultipleYieldCurveFinderDataBundle(DOUBLE_CURVE_INSTRUMENTS, SWAP_VALUES, null, unknownCurveNodes, unknownCurveInterpolators, true,
        FX_MATRIX);
    DOUBLE_CURVE_JACOBIAN_WITH_FD_INTERPOLATOR_SENSITIVITY = new MultipleYieldCurveFinderJacobian(data, PAR_RATE_SENSITIVITY_CALCULATOR);

    // There is at least one other root in the double curve finding problem (i.e. two curves that give the correct par rates but are different from the curves used to generate the par rates in the
    // first place),
    // Starting close to the solution ensures we get the right one
    final double[] rates = new double[TIME_GRID.length];
    for (int i = 0; i < FUNDING_YIELDS.length; i++) {
      rates[i] = FUNDING_YIELDS[i] + 0.02;
    }

    for (int i = 0; i < LIBOR_YIELDS.length; i++) {
      rates[i + FUNDING_YIELDS.length] = LIBOR_YIELDS[i] + 0.02;
    }
    X0 = new DoubleMatrix1D(rates);
  }

	

  

	

  

	

  

	

  

	

  

	

  

	

  

	

  

	

  

	

  

	

  

	

  

	

  

	

  

	

  

	

  

	

  /**
   * 
   * @param payments
   * @param fundingCurveName
   * @param liborCurveName
   * @return
   */
  

	

  

	

  static {
    final int[] payments = new int[] {1, 2, 3, 4, 6, 8, 10, 14, 20, 30, 40, 50, 60 };

    FUNDING_CURVE_TIMES = new double[] {1, 2, 5, 10, 20, 31 };
    LIBOR_CURVE_TIMES = new double[] {0.5, 1, 2, 5, 10, 20, 31 };
    FUNDING_YIELDS = new double[] {0.021, 0.036, 0.06, 0.054, 0.049, 0.044 };
    LIBOR_YIELDS = new double[] {0.01, 0.02, 0.035, 0.06, 0.055, 0.05, 0.045 };

    final int n = payments.length;
    TIME_GRID = new double[n];
    SWAP_VALUES = new double[n];
    SINGLE_CURVE_INSTRUMENTS = new ArrayList<InstrumentDerivative>();
    DOUBLE_CURVE_INSTRUMENTS = new ArrayList<InstrumentDerivative>();
    EXTRAPOLATOR = CombinedInterpolatorExtrapolatorFactory.getInterpolator(Interpolator1DFactory.DOUBLE_QUADRATIC, LINEAR_EXTRAPOLATOR, FLAT_EXTRAPOLATOR);

    final YieldCurveBundle curveBundle = new YieldCurveBundle();
    FUNDING_CURVE = makeYieldCurve(FUNDING_YIELDS, FUNDING_CURVE_TIMES, EXTRAPOLATOR);
    LIBOR_CURVE = makeYieldCurve(LIBOR_YIELDS, LIBOR_CURVE_TIMES, EXTRAPOLATOR);
    curveBundle.setCurve(FUNDING_CURVE_NAME, FUNDING_CURVE);
    curveBundle.setCurve(LIBOR_CURVE_NAME, LIBOR_CURVE);

    InstrumentDerivative instrument;
    double swapRate;
    for (int i = 0; i < n; i++) {
      instrument = setupSwap(payments[i], FUNDING_CURVE_NAME, LIBOR_CURVE_NAME);
      swapRate = instrument.accept(SWAP_RATE_CALCULATOR, curveBundle);
      instrument = setParSwapRate((FixedFloatSwap) instrument, swapRate);

      DOUBLE_CURVE_INSTRUMENTS.add(instrument);
      instrument = setupSwap(payments[i], LIBOR_CURVE_NAME, LIBOR_CURVE_NAME);
      instrument = setParSwapRate((FixedFloatSwap) instrument, swapRate);
      SINGLE_CURVE_INSTRUMENTS.add(instrument);
      TIME_GRID[i] = payments[i] * 0.5;
      SWAP_VALUES[i] = swapRate;
    }

    LinkedHashMap<String, Interpolator1D> unknownCurveInterpolators = new LinkedHashMap<String, Interpolator1D>();
    LinkedHashMap<String, double[]> unknownCurveNodes = new LinkedHashMap<String, double[]>();
    unknownCurveInterpolators.put(LIBOR_CURVE_NAME, EXTRAPOLATOR);
    unknownCurveNodes.put(LIBOR_CURVE_NAME, TIME_GRID);
    MultipleYieldCurveFinderDataBundle data = new MultipleYieldCurveFinderDataBundle(SINGLE_CURVE_INSTRUMENTS, SWAP_VALUES, null, unknownCurveNodes,
        unknownCurveInterpolators, false, FX_MATRIX);
    SINGLE_CURVE_FINDER = new MultipleYieldCurveFinderFunction(data, PAR_RATE_CALCULATOR);
    SINGLE_CURVE_JACOBIAN = new MultipleYieldCurveFinderJacobian(data, PAR_RATE_SENSITIVITY_CALCULATOR);

    data = new MultipleYieldCurveFinderDataBundle(SINGLE_CURVE_INSTRUMENTS, new double[SINGLE_CURVE_INSTRUMENTS.size()], null, unknownCurveNodes,
        unknownCurveInterpolators, true, FX_MATRIX);
    SINGLE_CURVE_JACOBIAN_WITH_FD_INTERPOLATOR_SENSITIVITY = new MultipleYieldCurveFinderJacobian(data, PAR_RATE_SENSITIVITY_CALCULATOR);

    unknownCurveInterpolators = new LinkedHashMap<String, Interpolator1D>();
    unknownCurveNodes = new LinkedHashMap<String, double[]>();
    unknownCurveInterpolators.put(FUNDING_CURVE_NAME, EXTRAPOLATOR);
    unknownCurveInterpolators.put(LIBOR_CURVE_NAME, EXTRAPOLATOR);
    unknownCurveNodes.put(FUNDING_CURVE_NAME, FUNDING_CURVE_TIMES);
    unknownCurveNodes.put(LIBOR_CURVE_NAME, LIBOR_CURVE_TIMES);
    data = new MultipleYieldCurveFinderDataBundle(DOUBLE_CURVE_INSTRUMENTS, SWAP_VALUES, null, unknownCurveNodes, unknownCurveInterpolators, false,
        FX_MATRIX);

    DOUBLE_CURVE_FINDER = new MultipleYieldCurveFinderFunction(data, PAR_RATE_CALCULATOR);
    DOUBLE_CURVE_JACOBIAN = new MultipleYieldCurveFinderJacobian(data, PAR_RATE_SENSITIVITY_CALCULATOR);

    data = new MultipleYieldCurveFinderDataBundle(DOUBLE_CURVE_INSTRUMENTS, SWAP_VALUES, null, unknownCurveNodes, unknownCurveInterpolators, true,
        FX_MATRIX);
    DOUBLE_CURVE_JACOBIAN_WITH_FD_INTERPOLATOR_SENSITIVITY = new MultipleYieldCurveFinderJacobian(data, PAR_RATE_SENSITIVITY_CALCULATOR);

    // There is at least one other root in the double curve finding problem (i.e. two curves that give the correct par rates but are different from the curves used to generate the par rates in the
    // first place),
    // Starting close to the solution ensures we get the right one
    final double[] rates = new double[TIME_GRID.length];
    for (int i = 0; i < FUNDING_YIELDS.length; i++) {
      rates[i] = FUNDING_YIELDS[i] + 0.02;
    }

    for (int i = 0; i < LIBOR_YIELDS.length; i++) {
      rates[i + FUNDING_YIELDS.length] = LIBOR_YIELDS[i] + 0.02;
    }
    X0 = new DoubleMatrix1D(rates);
  }


}
