/**
 * Copyright (C) 2009 - present by OpenGamma Inc. and the OpenGamma group of companies
 *
 * Please see distribution for license.
 */
package com.opengamma.engine.view.calc; 

import static com.opengamma.util.functional.Functional.submapByKeySet; 

import java.io.File; 
import java.io.FileWriter; 
import java.io.IOException; 
import java.util.ArrayList; 
import java.util.Collection; 
import java.util.Collections; 
import java.util.HashMap; 
import java.util.HashSet; 
import java.util.Iterator; 
import java.util.LinkedHashSet; 
import java.util.LinkedList; 
import java.util.List; 
import java.util.Map; 
import java.util.Set; 
import java.util.TreeMap; 
import java.util.concurrent.BlockingQueue; 
import java.util.concurrent.ConcurrentHashMap; 
import java.util.concurrent.ExecutionException; 
import java.util.concurrent.ExecutorService; 
import java.util.concurrent.Future; 
import java.util.concurrent.LinkedBlockingQueue; 
import java.util.concurrent.TimeUnit; 
import java.util.concurrent.TimeoutException; 

import org.slf4j.Logger; 
import org.slf4j.LoggerFactory; 
import org.threeten.bp.Duration; 
import org.threeten.bp.Instant; 

import com.opengamma.DataNotFoundException; 
import com.opengamma.OpenGammaRuntimeException; 
import com.opengamma.engine.depgraph.DependencyGraph; 
import com.opengamma.engine.depgraph.DependencyNode; 
import com.opengamma.engine.depgraph.DependencyNodeFilter; 
import com.opengamma.engine.function.MarketDataSourcingFunction; 
import com.opengamma.engine.function.blacklist.FunctionBlacklistQuery; 
import com.opengamma.engine.marketdata.MarketDataSnapshot; 
import com.opengamma.engine.marketdata.OverrideOperation; 
import com.opengamma.engine.target.ComputationTargetType; 
import com.opengamma.engine.value.ComputedValue; 
import com.opengamma.engine.value.ComputedValueResult; 
import com.opengamma.engine.value.ValueRequirement; 
import com.opengamma.engine.value.ValueSpecification; 
import com.opengamma.engine.view.AggregatedExecutionLog; 
import com.opengamma.engine.view.ExecutionLog; 
import com.opengamma.engine.view.ExecutionLogMode; 
import com.opengamma.engine.view.ExecutionLogModeSource; 
import com.opengamma.engine.view.ExecutionLogWithContext; 
import com.opengamma.engine.view.InMemoryViewComputationResultModel; 
import com.opengamma.engine.view.ViewCalculationConfiguration; 
import com.opengamma.engine.view.ViewComputationResultModel; 
import com.opengamma.engine.view.ViewDefinition; 
import com.opengamma.engine.view.ViewProcessContext; 
import com.opengamma.engine.view.cache.CacheSelectHint; 
import com.opengamma.engine.view.cache.MissingMarketDataSentinel; 
import com.opengamma.engine.view.cache.NotCalculatedSentinel; 
import com.opengamma.engine.view.cache.ViewComputationCache; 
import com.opengamma.engine.view.calc.stats.GraphExecutorStatisticsGatherer; 
import com.opengamma.engine.view.calcnode.CalculationJobResultItem; 
import com.opengamma.engine.view.calcnode.MissingInput; 
import com.opengamma.engine.view.calcnode.MutableExecutionLog; 
import com.opengamma.engine.view.compilation.CompiledViewDefinitionWithGraphsImpl; 
import com.opengamma.engine.view.execution.ViewCycleExecutionOptions; 
import com.opengamma.engine.view.listener.ComputationResultListener; 
import com.opengamma.id.UniqueId; 
import com.opengamma.id.VersionCorrection; 
import com.opengamma.util.ArgumentChecker; 
import com.opengamma.util.log.LogLevel; 
import com.opengamma.util.log.SimpleLogEvent; 
import com.opengamma.util.tuple.Pair; 
import com.opengamma.lambdava.tuple.Pair; 

import static com.opengamma.lambdava.streams.Lambdava.submapByKeySet; 

/**
 * Holds all data and actions for a single computation pass. The view cycle may be executed at most once.
 * <p>
 * The cycle is thread-safe for readers, for example obtaining the current state or the result, but is only designed for a single executor.
 */
    SingleComputationCycle   {
	

  

	

  /**
   * The default property used to manipulate all root market data prior to cycle execution.
   */
  

	

   enum  NodeStateFlag {
    /**
     * Node was executed successfully.
     */
     , 
    /**
     * Node was executed but failed.
     */
     , 
    /**
     * Node was not executed because of function blacklist suppression.
     */
    ;}

	;

	

  // Injected inputs
  

	
  

	
  

	
  

	
  

	
  

	
  

	

  

	
  

	
  

	

  

	

  

	
  

	

  

	
  

	
  

	

  // Output
  

	

  

	

  

	

  //-------------------------------------------------------------------------
  

	

  

	

  

	

  /**
   * Gets the start time
   *
   * @return the start time
   */
  

	

  /**
   * Gets the end time.
   *
   * @return the end time
   */
  

	

  /**
   * @return the viewDefinition
   */
  

	

  

	

  

	

  

	

  

	

  

	

  //-------------------------------------------------------------------------
  

	

  

	

  

	

  

	

  //-------------------------------------------------------------------------
  

	

  

	

  

	

  

	

  

	

  

	

  

	

  

	

  //--------------------------------------------------------------------------
  // REVIEW jonathan 2011-03-18 -- The following comment should be given some sort of 'listed' status for preservation :-)
  // REVIEW kirk 2009-11-03 -- This is a database kernel. Act accordingly.

  /**
   * Synchronously runs the cycle.
   *
   * @param previousCycle the previous cycle from which a delta cycle should be performed, or null to perform a full cycle
   * @param marketDataSnapshot the market data snapshot with which to execute the cycle, not null
   * @param calcJobResultExecutorService the executor to use for streaming calculation job result consumption, not null
   * @throws InterruptedException if the thread is interrupted while waiting for the computation cycle to complete. Execution of any outstanding jobs will be cancelled, but {@link #release()} still
   *           must be called.
   */
  

	

  /**
   * Creates a map containing the "shift" operations to apply to market data or each calculation configuration. If there is no operation to apply, the map contains null for that configuration.
   *
   * @return the map of computation cache to shift operations
   */
  

	

  

	

  

	

  

	

  /**
   * Ensures that a computation cache exists for for each calculation configuration.
   */
  

	

  

	

  

	

  /**
   * Determine which live data inputs have changed between iterations, and:
   * <ul>
   * <li>Copy over all values that can be demonstrated to be the same from the previous iteration (because no input has changed)
   * <li>Only recompute the values that could have changed based on live data inputs
   * </ul>
   *
   * @param previousCycle Previous iteration. It must not have been cleaned yet ({@link #releaseResources()}).
   */
  

	

  

	

  //-------------------------------------------------------------------------
  /**
   * Processes the results from one or more calculation jobs.
   * <p>
   * This generates a fragment result immediately from the given results, and also adds the values to the full result
   * model.
   *
   * @param results  the execution results, not null
   */
  /*package*/ 

	

  

	

  

	

  

	

  

	

  

	

  //-------------------------------------------------------------------------
  /**
   * Returns the dependency graph used by this cycle for the given calculation configuration.
   *
   * @param calcConfName calculation configuration name
   * @return the dependency graph
   */
  

	

  /**
   * Creates a subset of the dependency graph for execution. This will only include nodes that do are not dummy ones to source market data, have been considered executed by a delta from the previous
   * cycle, or are being suppressed by the execution blacklist. Note that this will update the cache with synthetic output values from suppressed nodes and alter the execution state of any nodes not
   * in the resultant subgraph.
   *
   * @param calcConfName calculation configuration name
   * @return a dependency graph with any nodes which have already been satisfied filtered out, not null See {@link #computeDelta} and how it calls {@link #markExecuted}.
   */
  

	

  //--------------------------------------------------------------------------
  

	

  

	

  

	

  

	

  

	

  

	

  

	

  


}
