/**
 * Copyright (C) 2009 - present by OpenGamma Inc. and the OpenGamma group of companies
 *
 * Please see distribution for license.
 */
package com.opengamma.engine.view.calc; 

import java.text.MessageFormat; 
import java.util.ArrayList; 
import java.util.Collection; 
import java.util.Collections; 
import java.util.HashMap; 
import java.util.HashSet; 
import java.util.Iterator; 
import java.util.List; 
import java.util.Map; 
import java.util.Set; 
import java.util.concurrent.ConcurrentHashMap; 
import java.util.concurrent.ConcurrentMap; 
import java.util.concurrent.CountDownLatch; 
import java.util.concurrent.ExecutorService; 
import java.util.concurrent.Executors; 
import java.util.concurrent.Future; 
import java.util.concurrent.TimeUnit; 

import org.slf4j.Logger; 
import org.slf4j.LoggerFactory; 
import org.threeten.bp.Duration; 
import org.threeten.bp.Instant; 

import com.google.common.collect.Maps; 
import com.google.common.collect.Sets; 
import com.opengamma.DataNotFoundException; 
import com.opengamma.OpenGammaRuntimeException; 
import com.opengamma.core.change.ChangeEvent; 
import com.opengamma.core.change.ChangeListener; 
import com.opengamma.engine.ComputationTargetSpecification; 
import com.opengamma.engine.depgraph.DependencyGraph; 
import com.opengamma.engine.depgraph.DependencyNode; 
import com.opengamma.engine.depgraph.DependencyNodeFilter; 
import com.opengamma.engine.marketdata.MarketDataListener; 
import com.opengamma.engine.marketdata.MarketDataSnapshot; 
import com.opengamma.engine.marketdata.availability.MarketDataAvailabilityProvider; 
import com.opengamma.engine.marketdata.spec.MarketDataSpecification; 
import com.opengamma.engine.target.ComputationTargetReference; 
import com.opengamma.engine.value.ValueRequirement; 
import com.opengamma.engine.value.ValueSpecification; 
import com.opengamma.engine.view.ViewCalculationConfiguration; 
import com.opengamma.engine.view.ViewComputationResultModel; 
import com.opengamma.engine.view.ViewDefinition; 
import com.opengamma.engine.view.ViewProcessContext; 
import com.opengamma.engine.view.ViewProcessImpl; 
import com.opengamma.engine.view.calc.trigger.CombinedViewCycleTrigger; 
import com.opengamma.engine.view.calc.trigger.FixedTimeTrigger; 
import com.opengamma.engine.view.calc.trigger.RecomputationPeriodTrigger; 
import com.opengamma.engine.view.calc.trigger.RunAsFastAsPossibleTrigger; 
import com.opengamma.engine.view.calc.trigger.SuccessiveDeltaLimitTrigger; 
import com.opengamma.engine.view.calc.trigger.ViewCycleEligibility; 
import com.opengamma.engine.view.calc.trigger.ViewCycleTrigger; 
import com.opengamma.engine.view.calc.trigger.ViewCycleTriggerResult; 
import com.opengamma.engine.view.calc.trigger.ViewCycleType; 
import com.opengamma.engine.view.compilation.CompiledViewDefinitionWithGraphsImpl; 
import com.opengamma.engine.view.compilation.ViewCompilationServices; 
import com.opengamma.engine.view.compilation.ViewDefinitionCompiler; 
import com.opengamma.engine.view.execution.ViewCycleExecutionOptions; 
import com.opengamma.engine.view.execution.ViewExecutionFlags; 
import com.opengamma.engine.view.execution.ViewExecutionOptions; 
import com.opengamma.engine.view.listener.ComputationResultListener; 
import com.opengamma.id.ObjectId; 
import com.opengamma.id.UniqueId; 
import com.opengamma.id.VersionCorrection; 
import com.opengamma.util.ArgumentChecker; 
import com.opengamma.util.TerminatableJob; 
import com.opengamma.util.monitor.OperationTimer; 
import com.opengamma.util.tuple.Pair; 
import com.opengamma.lambdava.tuple.Pair; 

/**
 * The job which schedules and executes computation cycles for a view process.
 */
    ViewComputationJob  extends TerminatableJob   {
	
  

	

  

	
  

	

  

	
  

	
  

	
  

	
  

	
  

	
  

	
  

	

  

	
  

	

  

	
  

	
  

	
  

	
  

	

  

	
  

	
  

	

  

	
  

	
  

	
  

	
  

	
  

	

  /**
   * Nanoseconds
   */
  

	

  

	

  

	

  

	

  // TODO get rid of the private getters
  

	

  

	

  

	

  

	

  

	

  

	

  /**
   * Determines whether to run, and runs if required, a single computation cycle using the following rules:
   * <ul>
   * <li>A computation cycle can only be triggered if the relevant minimum computation period has passed since the start of the previous cycle.
   * <li>A computation cycle will be forced if the relevant maximum computation period has passed since the start of the previous cycle.
   * <li>A full computation is preferred over a delta computation if both are possible.
   * <li>Performing a full computation also updates the times to the next delta computation; i.e. a full computation is considered to be as good as a delta.
   * </ul>
   */
  

	

  

	

  

	

  

	

  

	

  

	

  

	

  

	

  

	

  

	

  

	

  

	

  /**
   * Indicates that the view definition itself has changed. It is not necessary to call {@link #dirtyCompilation()} as well.
   */
  

	

  /**
   * Indicates that changes have occurred which may affect the compilation, and the view definition should be recompiled at the earliest opportunity.
   */
  

	

  /**
   * Forces a cycle to run, regardless of how long has elapsed since the previous cycle.
   */
  

	

  /**
   * Requests a cycle to run, honouring minimum times between cycles.
   */
  

	

  

	

  

	

  

	

  

	

  

	

  /**
   * Returns the set of unique identifiers that were previously used as targets in the dependency graph for object identifiers (or external identifiers) that now resolve differently.
   *
   * @param previousResolutions the previous cycle's resolution of identifiers, not null
   * @param versionCorrection the resolver version correction for this cycle, not null
   * @return the invalid identifier set, or null if none are invalid
   */
  

	

  /**
   * Mark a set of nodes for inclusion (TRUE) or exclusion (FALSE) based on the filter. A node is included if the filter accepts it and all of its inputs are also marked for inclusion. A node is
   * excluded if the filter rejects it or any of its inputs are rejected. This will operate recursively, processing all nodes to the leaves of the graph.
   * <p>
   * The {@link DependencyGraph#subGraph} operation doesn't work for us as it can leave nodes in the sub-graph that have inputs that aren't in the graph. Invalid nodes identified by the filter need to
   * remove all the graph up to the terminal output root so that we can rebuild it.
   *
   * @param include the map to build the result into
   * @param nodes the nodes to process
   * @param filter the filter to apply to the nodes
   */
  

	

  

	

  /**
   * Maintain the previously used dependency graphs by applying a node filter that identifies invalid nodes that must be recalculated (implying everything dependent on them must also be rebuilt). The
   * first call will extract the previously compiled graphs, subsequent calls will update the structure invalidating more nodes and increasing the number of missing requirements.
   *
   * @param previousGraphs the previously used graphs as a map from calculation configuration name to the graph and the value requirements that need to be recalculated, not null
   * @param filter the filter to identify invalid nodes, not null
   */
  

	

  

	

  /**
   * Gets the cached compiled view definition which may be re-used in subsequent computation cycles.
   * <p>
   * External visibility for tests.
   *
   * @return the cached compiled view definition, or null if nothing is currently cached
   */
  

	

  

	

  /**
   * Replaces the cached compiled view definition.
   * <p>
   * External visibility for tests.
   *
   * @param latestCompiledViewDefinition the compiled view definition, may be null
   */
  

	

  /**
   * Gets the view definition currently in use by the computation job.
   *
   * @return the view definition, not null
   */
  

	

  

	

  

	

  

	

  

	

  

	

  

	

  

	

  //-------------------------------------------------------------------------
  

	

  

	

  

	

  

	

  //-------------------------------------------------------------------------
  

	

  

	

  

	

  


}
