/**
 * Copyright (C) 2009 - present by OpenGamma Inc. and the OpenGamma group of companies
 *
 * Please see distribution for license.
 */
package com.opengamma.engine.view.calcnode; 

import java.util.ArrayList; 
import java.util.Collection; 
import java.util.Iterator; 
import java.util.List; 
import java.util.Set; 
import java.util.concurrent.Callable; 
import java.util.concurrent.CancellationException; 
import java.util.concurrent.ConcurrentMap; 
import java.util.concurrent.ExecutorService; 
import java.util.concurrent.Future; 

import org.slf4j.Logger; 
import org.slf4j.LoggerFactory; 

import com.google.common.collect.MapMaker; 
import com.google.common.collect.Sets; 
import com.opengamma.engine.ComputationTarget; 
import com.opengamma.engine.ComputationTargetResolver; 
import com.opengamma.engine.ComputationTargetSpecification; 
import com.opengamma.engine.function.CompiledFunctionService; 
import com.opengamma.engine.function.FunctionExecutionContext; 
import com.opengamma.engine.function.FunctionInputs; 
import com.opengamma.engine.function.FunctionInputsImpl; 
import com.opengamma.engine.function.FunctionInvoker; 
import com.opengamma.engine.function.blacklist.DummyFunctionBlacklistMaintainer; 
import com.opengamma.engine.function.blacklist.DummyFunctionBlacklistQuery; 
import com.opengamma.engine.function.blacklist.FunctionBlacklistMaintainer; 
import com.opengamma.engine.function.blacklist.FunctionBlacklistQuery; 
import com.opengamma.engine.function.blacklist.FunctionBlacklistedException; 
import com.opengamma.engine.target.ComputationTargetResolverUtils; 
import com.opengamma.engine.target.lazy.LazyComputationTargetResolver; 
import com.opengamma.engine.value.ComputedValue; 
import com.opengamma.engine.value.ValueRequirement; 
import com.opengamma.engine.value.ValueSpecification; 
import com.opengamma.engine.view.ViewProcessor; 
import com.opengamma.engine.view.cache.DeferredViewComputationCache; 
import com.opengamma.engine.view.cache.DirectWriteViewComputationCache; 
import com.opengamma.engine.view.cache.NotCalculatedSentinel; 
import com.opengamma.engine.view.cache.ViewComputationCache; 
import com.opengamma.engine.view.cache.ViewComputationCacheSource; 
import com.opengamma.engine.view.cache.WriteBehindViewComputationCache; 
import com.opengamma.engine.view.calcnode.stats.FunctionInvocationStatisticsGatherer; 
import com.opengamma.util.ArgumentChecker; 
import com.opengamma.util.InetAddressUtils; 
import com.opengamma.util.async.AsynchronousExecution; 
import com.opengamma.util.async.AsynchronousOperation; 
import com.opengamma.util.async.AsynchronousResult; 
import com.opengamma.util.async.ResultListener; 
import com.opengamma.util.time.DateUtils; 
import com.opengamma.util.tuple.Pair; 
import com.opengamma.lambdava.tuple.Pair; 

/**
 * A calculation node implementation. The node can only be used by one thread - i.e. executeJob cannot be called concurrently to do multiple jobs. To execute multiple jobs concurrently separate
 * calculation nodes must be used.
 * <p>
 * The function repository (and anything else) must be properly initialized and ready for use by the node when it receives its first job. Responsibility for initialization should therefore lie with
 * whatever will logically be dispatching jobs. This is typically a {@link ViewProcessor} for local nodes or a {@link RemoteNodeClient} for remote nodes.
 */
    SimpleCalculationNode  extends SimpleCalculationNodeState   {
	

  /**
   * Interface used by deferred executions. Any {@link AsynchronousExecution} exceptions thrown by this class will supply either an instance of {@code Deferred} on the original method type, or the
   * original method type. Instances of {@code Deferred} are not bound (tightly) to their original calculation node instance allowing them to run on another (for example if the original node is now
   * being reused for another job). Thus the caller must supply the node that will now host the deferred operation. The state from the original node must be preserved and then restored into the new
   * host node.
   *
   * @param <T> the original return type of the asynchronous method
   */
      Deferred <T> {
		

    


	}

	

  

	
  

	
  

	

  

	
  

	

  /**
   * The deferred caches for performing write-behind and late write completion (asynchronous flush) to the value cache. The map contains weak values so that entries can be cleared when the deferred
   * cache is dead (i.e. no calculation node is currently using it and its writer is not running).
   */
  

	

  

	
  

	
  

	
  

	
  

	
  

	
  

	
  

	
  

	
  

	
  

	
  

	

  

	

  //-------------------------------------------------------------------------
  

	

  

	

  

	

  

	

  

	

  

	

  /**
   * Sets whether to use a write-behind strategy when writing to the shared value cache. Write-behind can work well if the cost of writing is high (e.g. network overhead). If the write is cheap (e.g.
   * to an in-process, in-memory store) then the overheads of the write-behind become a burden. An executor service must be available if this is selected.
   *
   * @param writeBehind true to use write-behind on the shared value cache, false not to
   */
  

	

  /**
   * Sets whether to use a write-behind strategy when writing to the private value cache. Write-behind can work well if the cost of writing is high (e.g. disk overhead). If the write is cheap (e.g. to
   * an in-process, in-memory store) then the overheads of the write-behind become a burden. An executor service must be available if this is selected.
   *
   * @param writeBehind true to use write-behind on the private value cache, false not to
   */
  

	

  

	

  

	

  

	

  

	

  /**
   * Sets a function blacklist to use for checking each job invocation. The blacklist(s) used may suppress particular functions on this node, the logical group of nodes or at all nodes.
   *
   * @param query the blacklist to query against, not null
   */
  

	

  /**
   * Returns the function blacklist used for checking each job invocation.
   *
   * @return the blacklist queried, not null
   */
  

	

  /**
   * Sets a maintenance policy for updating one or more blacklist(s) when a job item throws an exception. The application of a policy at this point will depend on the nature of the function library.
   * Issuing updates from the function's {@link FunctionInvoker} can allow more specific control over how a function failure should be reported.
   *
   * @param update the maintainer to notify of a job item that threw an exception
   */
  

	

  

	

  

	

  

	

  

	

  

	

  

	

  //-------------------------------------------------------------------------
  

	

  

	

  /**
   * Invokes all of the items from a calculation job on this node. If asynchronous execution occurs, the {@link AsynchronousExecution} will report either a {@code Deferred&lt;CalculationJobResult&gt;}
   * or a {@code CalculationJobResult}. This is to allow the node to be reused if calculations are running externally and for tail job executions on the local nodes to start immediately but block the
   * central dispatcher until the cache has been flushed.
   * <p>
   * If a deferred result is returned, and this node will be used for other work, its state must be saved and then later restored into a node that will be used to complete the deferred action.
   *
   * @param job the job to execute
   * @return the job result
   * @throws AsynchronousExecution if the job is completing asynchronously
   */
  

	

      ExecuteJobFinish   {
		

    

		

    

		

    


	}

	

  

	

  

	

  

	

  //-------------------------------------------------------------------------
  

	

  

	

  /**
   * Executes one or more items from the supplied iterator, populating the supplied list. If a job item starts running asynchronously, an exception will be thrown. At resumption of the operation,
   * another call to this method will occur with the same parameters allowing it to continue with the remaining items the iterator has.
   *
   * @param jobItemItr the job items to execute, not null
   * @param resultItems the list to populate with results, not null
   */
  

	

      ExecuteJobItemsInvoke {
		

    

		
    

		
    

		

    

		

    

		

    

		

    


	}

	

      ExecuteJobItemsResume   {
		

    

		
    

		
    

		
    

		

    

		

    


	}

	

  

	

      ExecuteJobItemsFinish   {
		

    

		
    

		

    

		

    


	}

	

  

	

  

	

  

	

  

	

  

	

  <<<<<<< E:\ssmerge\fstmerge_tmp1502139582173\fstmerge_var1_1609105373391422391.java
private void invoke(final CalculationJobItem jobItem, final DeferredInvocationStatistics statistics, final CalculationJobResultItemBuilder resultItemBuilder) throws AsynchronousExecution {
    final Set<ValueSpecification> outputs = jobItem.getOutputs();
    final String functionUniqueId = jobItem.getFunctionUniqueIdentifier();
    Future<ComputationTarget> targetFuture = null;
    ComputationTarget target = null;
    if (isUseAsynchronousTargetResolve()) {
      targetFuture = getExecutorService().submit(new Callable<ComputationTarget>() {
        @Override
        public ComputationTarget call() {
          return getTargetResolver().resolve(jobItem.getComputationTargetSpecification());
        }
      });
    } else {
      target = LazyComputationTargetResolver.resolve(getTargetResolver(), jobItem.getComputationTargetSpecification());
      if (target == null) {
        postEvaluationErrors(outputs, NotCalculatedSentinel.EVALUATION_ERROR);
        resultItemBuilder.withException(ERROR_CANT_RESOLVE, "Unable to resolve target " + jobItem.getComputationTargetSpecification());
        return;
      }
    }
    final FunctionInvoker invoker = getFunctions().getInvoker(functionUniqueId);
    if (invoker == null) {
      postEvaluationErrors(outputs, NotCalculatedSentinel.EVALUATION_ERROR);
      resultItemBuilder.withException(ERROR_BAD_FUNCTION, "Unable to locate " + functionUniqueId + " in function repository");
      return;
    }
    // set parameters
    getFunctionExecutionContext().setFunctionParameters(jobItem.getFunctionParameters());
    // assemble inputs
    final Collection<ValueSpecification> inputValueSpecs = jobItem.getInputs();
    final Collection<ComputedValue> inputs = new ArrayList<ComputedValue>(inputValueSpecs.size());
    final Set<ValueSpecification> missing = Sets.newHashSetWithExpectedSize(inputValueSpecs.size());
    int inputBytes = 0;
    int inputSamples = 0;
    final DeferredViewComputationCache cache = getCache();
    for (final Pair<ValueSpecification, Object> input : cache.getValues(inputValueSpecs, getJob().getCacheSelectHint())) {
      if ((input._2() == null) || (input._2() instanceof MissingInput)) {
        missing.add(input._1());
      } else {
        final ComputedValue value = new ComputedValue(input._1(), input._2());
        inputs.add(value);
        final Integer bytes = cache.estimateValueSize(value);
        if (bytes != null) {
          inputBytes += bytes;
          inputSamples++;
        }
      }
    }
    statistics.setDataInputBytes(inputBytes, inputSamples);
    if (!missing.isEmpty()) {
      if (invoker.canHandleMissingInputs()) {
        s_logger.debug("Executing even with missing inputs {}", missing);
        resultItemBuilder.withPartialInputs(missing);
      } else {
        s_logger.info("Not able to execute as missing inputs {}", missing);
        if (targetFuture != null) {
          // Cancelling doesn't do anything so we have to block and clear the result
          try {
            targetFuture.get();
          } catch (final Throwable t) {
            s_logger.warn("Error resolving target", t);
            postEvaluationErrors(outputs, NotCalculatedSentinel.EVALUATION_ERROR);
            resultItemBuilder.withException(t);
            return;
          }
        }
        postEvaluationErrors(jobItem.getOutputs(), NotCalculatedSentinel.MISSING_INPUTS);
        resultItemBuilder.withMissingInputs(missing);
        return;
      }
    }
    final FunctionInputs functionInputs = new FunctionInputsImpl(getTargetResolver().getSpecificationResolver(), inputs, missing);
    if (target == null) {
      try {
        target = targetFuture.get();
      } catch (final Throwable t) {
        s_logger.warn("Error resolving target", t);
        postEvaluationErrors(outputs, NotCalculatedSentinel.EVALUATION_ERROR);
        resultItemBuilder.withException(t);
        return;
      }
      if (target == null) {
        postEvaluationErrors(outputs, NotCalculatedSentinel.EVALUATION_ERROR);
        resultItemBuilder.withException(ERROR_CANT_RESOLVE, "Unable to resolve target " + jobItem.getComputationTargetSpecification());
        return;
      }
    }
    // Execute
    statistics.beginInvocation(functionUniqueId);
    Set<ComputedValue> result;
    try {
      result = invoker.execute(getFunctionExecutionContext(), functionInputs, target, plat2290(outputs));
    } catch (final AsynchronousExecution e) {
      s_logger.debug("Asynchronous execution of {} at {}", jobItem, _nodeId);
      final AsynchronousOperation<Deferred<Void>> async = deferredOperation();
      e.setResultListener(new ResultListener<Set<ComputedValue>>() {
        @Override
        public void operationComplete(final AsynchronousResult<Set<ComputedValue>> result) {
          s_logger.debug("Job item {} result available", jobItem);
          async.getCallback().setResult(new Deferred<Void>() {
            @Override
            public Void call(final SimpleCalculationNode self) {
              s_logger.debug("Asynchronous result for {} at {}", jobItem, self._nodeId);
              Set<ComputedValue> results;
              try {
                results = result.getResult();
              } catch (final FunctionBlacklistedException e) {
                self.invocationBlacklisted(jobItem, resultItemBuilder);
                return null;
              } catch (final Throwable t) {
                self.invokeException(outputs, t, resultItemBuilder);
                return null;
              }
              self.invokeResult(invoker, statistics, missing, outputs, results, resultItemBuilder);
              return null;
            }
          });
        }
      });
      async.getResult().call(this);
      return;
    } catch (final FunctionBlacklistedException e) {
      invocationBlacklisted(jobItem, resultItemBuilder);
      return;
    } catch (final Throwable t) {
      invokeException(outputs, t, resultItemBuilder);
      return;
    }
    invokeResult(invoker, statistics, missing, outputs, result, resultItemBuilder);
  }
=======
>>>>>>> E:\ssmerge\fstmerge_tmp1502139582173\fstmerge_var2_5795055190574497308.java



}
